from web3.auto import w3
from eth_account.messages import encode_defunct
from eth_account.datastructures import SignedMessage
from web3 import Web3


def sign_msg(msg: str, private_key: str) -> str:
    message = encode_defunct(text=msg)
    signed_message: SignedMessage = w3.eth.account.sign_message(
        message, private_key=private_key)
    return signed_message.signature.hex()


def get_key_seed(signature: int) -> bytes:
    key_seed = Web3.keccak(text=str(int(hex(signature), 16)))
    # key_seed = Web3.solidityKeccak(['uint256'], [signature])
    return key_seed



from typing import Tuple
from sympy.core.numbers import igcdex

# A type that represents a point (x,y) on an elliptic curve.
ECPoint = Tuple[int, int]


def div_mod(n: int, m: int, p: int) -> int:
    """
    Finds a nonnegative integer 0 <= x < p such that (m * x) % p == n
    """
    a, b, c = igcdex(m, p)
    assert c == 1
    return (n * a) % p


def ec_add(point1: ECPoint, point2: ECPoint, p: int) -> ECPoint:
    """
    Gets two points on an elliptic curve mod p and returns their sum.
    Assumes the points are given in affine form (x, y) and have different x coordinates.
    """
    assert (point1[0] - point2[0]) % p != 0
    m = div_mod(point1[1] - point2[1], point1[0] - point2[0], p)
    x = (m * m - point1[0] - point2[0]) % p
    y = (m * (point1[0] - x) - point1[1]) % p
    return x, y


def ec_double(point: ECPoint, alpha: int, p: int) -> ECPoint:
    """
    Doubles a point on an elliptic curve with the equation y^2 = x^3 + alpha*x + beta mod p.
    Assumes the point is given in affine form (x, y) and has y != 0.
    """
    assert point[1] % p != 0
    m = div_mod(3 * point[0] * point[0] + alpha, 2 * point[1], p)
    x = (m * m - 2 * point[0]) % p
    y = (m * (point[0] - x) - point[1]) % p
    return x, y


def ec_mult(m: int, point: ECPoint, alpha: int, p: int) -> ECPoint:
    """
    Multiplies by m a point on the elliptic curve with equation y^2 = x^3 + alpha*x + beta mod p.
    Assumes the point is given in affine form (x, y) and that 0 < m < order(point).
    """
    if m == 1:
        return point
    if m % 2 == 0:
        return ec_mult(m // 2, ec_double(point, alpha, p), alpha, p)
    return ec_add(ec_mult(m - 1, point, alpha, p), point, p)



from .bin.signature import sign, get_stark_key_pair_from_signature
from .bin.blockchain_utils import sign_msg
from .typings import CreateOrderNoncePayload, CreateNewOrderBody, CoinStatPayload, StarkSignature
from .exception import CoinNotFoundError
from .constants import Config, MAX_INT_ALLOWANCE
from typing import Literal
from web3 import Web3, Account
import os

def params_to_dict(dict: dict) -> dict:
    del dict['self']
    d = {}
    for key, val in dict.items():
        if (val):
            d[key] = val
    return d


def sign_order_with_stark_private_key(stark_private_key: str, nonce: CreateOrderNoncePayload,):
    r, s = sign(msg_hash=int(nonce['msg_hash'], 16),
                stark_private_key=stark_private_key)

    create_order_request_data: CreateNewOrderBody = {
        "nonce": nonce['nonce'],
        "msg_hash": nonce['msg_hash'],
        "signature": {
            "r": hex(r),
            "s": hex(s)
        }
    }
    return create_order_request_data

def filter_ethereum_coin(coin_stats_payload: CoinStatPayload, coin: str):
    for coin_name, coin_stat in coin_stats_payload.items():
        if coin_stat["symbol"] == coin:
            return coin_stat
    raise CoinNotFoundError(f"Coin '{coin}' not found")

def get_nonce(signer, provider: Web3):
    base_nonce = provider.eth.get_transaction_count(signer.address)
    nonce_offset = 0
    return base_nonce + nonce_offset

def get_0x0_to_0x(address: str):
    if address and (address[:3] == '0x0' or address[:3] == '0X0'):
        return '0x' + address[4:]
    else:
        return address

def dequantize(number: float, decimals: int):
    factor = 10**decimals
    return number/factor

def get_allowance(user_address: str, stark_contract: str, token_contract: str, decimal: str, w3: Web3):
    contract_instance = w3.eth.contract(address=token_contract, abi=Config.ERC20_ABI) # type:ignore
    allowance = contract_instance.functions.allowance(user_address, stark_contract).call()
    allowance_decimal = dequantize(int(allowance), int(decimal))
    return allowance_decimal

def approve_unlimited_allowance_util(contract_address: str, token_contract: str, signer: Account, w3: Web3):
    gas_price = w3.eth.gas_price
    contract_instance = w3.eth.contract(address=token_contract, abi=Config.ERC20_ABI) # type:ignore

    gas_limit = contract_instance.functions.approve(
        contract_address,
        w3.to_int(int(MAX_INT_ALLOWANCE))
    ).estimateGas({"from": token_contract})
    # so basically, if token contract is not provided, it'll try to approve from 0 address account,
    # so providing a from account, other way is to initialize web3 using private key
    overrides = {
        'gas': gas_limit,
        'gasPrice': gas_price,
        'nonce': get_nonce(signer=signer, provider=w3),
    }
    amount = int(MAX_INT_ALLOWANCE)
    transaction_pre_build = contract_instance.functions.approve(contract_address, amount)
    transaction = transaction_pre_build.buildTransaction(overrides) # type: ignore
    signed_tx = signer.sign_transaction(transaction)
    # send this signed transaction to blockchain
    w3.eth.sendRawTransaction(signed_tx.rawTransaction).hex()
    approval = signed_tx
    return approval

def sign_withdrawal_tx_msg_hash(key_pair: dict, msg_hash: str):
    r, s = sign(msg_hash=int(msg_hash), stark_private_key=key_pair['stark_private_key'])
    signature = {
        'r': f"{hex(r)}",
        's': f"{hex(s)}",
        'recoveryParam': 0
    }
    return signature

def format_withdrawal_amount(amount: int, decimals: int, symbol: str):
    if symbol == 'eth':
        return str(Web3().fromWei(amount, 'ether')) if amount else '0'
    else:
        return str(dequantize(number=amount, decimals=decimals))

def sign_internal_tx_msg_hash(key_pair: dict, msg_hash: str):
    r, s = sign(int(msg_hash, 16), key_pair['stark_private_key'])
    signature: StarkSignature = {
        'r': hex(r),
        's': hex(s),
    } # type:ignore
    return signature

def filter_cross_chain_coin(config, coin, type):
    allowed_tokens = config['tokens']
    allowed_tokens_for_deposit = config['allowed_tokens_for_deposit']
    allowed_tokens_for_fast_withdrawal = config['allowed_tokens_for_fast_wd']

    if type == 'TOKENS':
        allowed_token = allowed_tokens[coin]
    elif type == 'DEPOSIT':
        allowed_token = next((token for token in allowed_tokens_for_deposit if token == coin), None)
    elif type == 'WITHDRAWAL':
        allowed_token = next((token for token in allowed_tokens_for_fast_withdrawal if token == coin), None)
    else:
        raise CoinNotFoundError('Type not found')
    if not allowed_token:
        raise CoinNotFoundError(f'Coin {coin} not found')

    current_coin = allowed_tokens[coin]
    return current_coin




class AuthenticationError(Exception):
    def __init__(self, message):
        super().__init__(message)
        self.name = 'AuthenticationError'

class ConnectNotCalled(Exception):
    pass

class CoinNotFoundError(Exception):
    def __init__(self, message):
        super().__init__(message)
        self.name = 'CoinNotFoundError'


class InvalidAmountError(Exception):
    def __init__(self, message):
        super().__init__(message)
        self.name = 'InvalidAmountError'


class AllowanceTooLowError(Exception):
    def __init__(self, message):
        super().__init__(message)
        self.name = 'AllowanceTooLowError'


class BalanceTooLowError(Exception):
    def __init__(self, message):
        super().__init__(message)
        self.name = 'BalanceTooLowError'


def is_authentication_error(err):
    return isinstance(err, AuthenticationError)


__all__ = [
    "AuthenticationError",
    "is_authentication_error",
    "CoinNotFoundError",
    "AllowanceTooLowError",
    "BalanceTooLowError",
    "InvalidAmountError",
]



import math
import json
import os
from typing import Optional, Tuple
import hashlib
from . import blockchain_utils
from .math_utils import div_mod, ec_mult, ECPoint, ec_add
from ecdsa.rfc6979 import generate_k


PEDERSEN_HASH_POINT_FILENAME = os.path.join(
    os.path.dirname(__file__), 'pedersen_params.json')
PEDERSEN_PARAMS = json.load(open(PEDERSEN_HASH_POINT_FILENAME))
FIELD_PRIME = PEDERSEN_PARAMS['FIELD_PRIME']
N_ELEMENT_BITS_ECDSA = math.floor(math.log(FIELD_PRIME, 2))
EC_ORDER = PEDERSEN_PARAMS['EC_ORDER']
CONSTANT_POINTS = PEDERSEN_PARAMS['CONSTANT_POINTS']
EC_GEN = CONSTANT_POINTS[1]
ALPHA = PEDERSEN_PARAMS['ALPHA']

# A type for the digital signature.
ECSignature = Tuple[int, int]


def generate_k_rfc6979(msg_hash: int, stark_private_key: str, seed: Optional[int] = None) -> int:
    stark_private_key_int = int(stark_private_key, 16)
    # Pad the message hash, for consistency with the elliptic.js library.
    if 1 <= msg_hash.bit_length() % 8 <= 4 and msg_hash.bit_length() >= 248:
        # Only if we are one-nibble short:
        msg_hash *= 16

    if seed is None:
        extra_entropy = b''
    else:
        extra_entropy = seed.to_bytes(math.ceil(seed.bit_length() / 8), 'big')

    return generate_k(EC_ORDER, stark_private_key_int, hashlib.sha256,
                      msg_hash.to_bytes(math.ceil(msg_hash.bit_length() / 8), 'big'),
                      extra_entropy=extra_entropy)


def inv_mod_curve_size(x: int) -> int:
    return div_mod(1, x, EC_ORDER)


def sign(msg_hash: int, stark_private_key: str, seed: Optional[int] = None) -> ECSignature:
    stark_private_key_int = int(stark_private_key, 16)
    # Note: msg_hash must be smaller than 2**N_ELEMENT_BITS_ECDSA.
    # Message whose hash is >= 2**N_ELEMENT_BITS_ECDSA cannot be signed.
    # This happens with a very small probability.
    assert 0 <= msg_hash < 2**N_ELEMENT_BITS_ECDSA, 'Message not signable.'

    # Choose a valid k. In our version of ECDSA not every k value is valid,
    # and there is a negligible probability a drawn k cannot be used for signing.
    # This is why we have this loop.
    while True:
        k = generate_k_rfc6979(msg_hash, stark_private_key, seed)
        # Update seed for next iteration in case the value of k is bad.
        if seed is None:
            seed = 1
        else:
            seed += 1

        # Cannot fail because 0 < k < EC_ORDER and EC_ORDER is prime.
        x = ec_mult(k, EC_GEN, ALPHA, FIELD_PRIME)[0]

        # DIFF: in classic ECDSA, we take int(x) % n.
        r = int(x)
        if not (1 <= r < 2**N_ELEMENT_BITS_ECDSA):
            # Bad value. This fails with negligible probability.
            continue

        if (msg_hash + r * stark_private_key_int) % EC_ORDER == 0:
            # Bad value. This fails with negligible probability.
            continue

        w = div_mod(k, msg_hash + r * stark_private_key_int, EC_ORDER)
        if not (1 <= w < 2**N_ELEMENT_BITS_ECDSA):
            # Bad value. This fails with negligible probability.
            continue

        s = inv_mod_curve_size(w)
        return r, s


def stark_private_key_to_ec_point_on_stark_curve(stark_private_key: str) -> ECPoint:
    stark_private_key_int = int(stark_private_key, 16)
    assert 0 < stark_private_key_int < EC_ORDER
    return ec_mult(stark_private_key_int, EC_GEN, ALPHA, FIELD_PRIME)


def stark_private_key_to_stark_public_key(stark_private_key: str) -> str:
    return hex(stark_private_key_to_ec_point_on_stark_curve(stark_private_key)[0])


def get_stark_key_pair_from_signature(
        signature: str
) -> "dict[str, str]":
    '''
    Derive a STARK key pair deterministically from an Ethereum key.
    This is the function used by the Brine frontend to derive a user's
    STARK key pair in a way that is recoverable. Programmatic traders may
    optionally derive their STARK key pair in the same way.
    :param signature: mandatory
    :type signature: str
    '''
    signature_int = int(signature, 16)
    hashed_signature = blockchain_utils.get_key_seed(signature=signature_int)
    # grind key
    grinded_key = grind_key(key_seed=int(hashed_signature.hex(), 16), key_value_limit=EC_ORDER)
    stark_private_key = hex(grinded_key)
    stark_public_key = stark_private_key_to_stark_public_key(
        stark_private_key=stark_private_key
    )
    return {
        'stark_private_key': stark_private_key,
        'stark_public_key': stark_public_key
        # 'stark_public_key_y_coordinate': public_y,
    }


def grind_key(key_seed: int, key_value_limit: int) -> int:
    max_allowed_value = 2**256 - (2**256 % key_value_limit)
    current_index = 0

    def indexed_sha256(seed: int, index: int) -> int:
        def padded_hex(x: int) -> str:
            # Hex string should have an even number of characters to convert to bytes.
            hex_str = hex(x)[2:]
            return hex_str if len(hex_str) % 2 == 0 else "0" + hex_str

        digest = hashlib.sha256(bytes.fromhex(padded_hex(seed) + padded_hex(index))).hexdigest()
        return int(digest, 16)

    key = indexed_sha256(seed=key_seed, index=current_index)
    while key >= max_allowed_value:
        current_index += 1
        key = indexed_sha256(seed=key_seed, index=current_index)

    return key % key_value_limit



import requests
from typing import Callable, Union, Optional
from .typings import LoginResponse


class Session(requests.Session):
    def __init__(self, refresh_tokens: Callable, url_base=None, *args, **kwargs):
        super(Session, self).__init__(*args, **kwargs)
        self.url_base = url_base
        self.refresh_tokens = refresh_tokens

    def request(self, method, url, **kwargs):
        modified_url = self.url_base + url
        response = super(Session, self).request(method, modified_url, **kwargs)

        if response.status_code == 401 and self.headers.get('Authorization') and response.json().get('payload').get('token_type') == 'access':
            del self.headers['Authorization']
            res = self.refresh_tokens()

            if res:
                self.headers.update(
                    {'Authorization': f"JWT {res['payload']['access']}"})
                response.headers.update(
                    {'Authorization': f"JWT {res['payload']['access']}"})

            return super(Session, self).request(method, modified_url, **kwargs)

        response.raise_for_status()
        return response



from .session import Session
from .utils import (
    params_to_dict,
    filter_ethereum_coin,
    get_nonce,
    get_0x0_to_0x,
    get_allowance,
    approve_unlimited_allowance_util,
    format_withdrawal_amount,
    sign_withdrawal_tx_msg_hash,
    filter_cross_chain_coin,
    sign_internal_tx_msg_hash,
    filter_cross_chain_coin,
    sign_order_with_stark_private_key
)
from .bin.blockchain_utils import sign_msg
from .exception import *
from typing import Optional, Union, List, Literal
from .typings import (
    Response,
    LoginResponse,
    FullDayPricePayload,
    CandleStickPayload,
    OrderBookPayload,
    RecentTradesPayload,
    ProfileInformationPayload,
    Balance,
    ProfitAndLossPayload,
    CreateOrderNoncePayload,
    CreateNewOrderBody,
    OrderPayload,
    CreateOrderNonceBody,
    CancelOrder,
    TradePayload,
    Order,
    TokenType,
    ListDepositParams,
    ListWithdrawalParams,
    InternalTransferInitiateBody,
    InternalTransferProcessBody,
    ListInternalTransferParams,
    InitiateWithdrawalPayload,
    ProcessFastWithdrawalPayload,
)
from web3 import Web3, Account
from .constants import Config
from web3.middleware.geth_poa import geth_poa_middleware

class Client:
    def __init__(self, option: Literal['mainnet', 'testnet'] = 'mainnet'):
        base_url = "https://api-testnet.tanx.fi" if option == 'testnet' else 'https://api.tanx.fi'
        self.session = Session(self.refresh_tokens, base_url)
        self.access_token: Optional[str] = None
        self.refresh_token: Optional[str] = None
        self.option = option

    def refresh_tokens(self, refresh_token: Optional[str] = None) -> Union[Response[TokenType], None]:
        if refresh_token or self.refresh_token:
            r = self.session.post('/sapi/v1/auth/token/refresh/', {
                "refresh": refresh_token if refresh_token else self.refresh_token
            })
            self.set_access_token(r.json()['payload']['access'])
            self.set_refresh_token(r.json()['payload']['refresh'])
            return r.json()

    def set_access_token(self, token: str) -> None:
        self.access_token = token
        self.session.headers.update(
            {'Authorization': f"JWT {token}"})
        
    def set_refresh_token(self, token: str) -> None:
        self.refresh_token = token

    def logout(self) -> None:
        self.access_token = None
        self.refresh_token = None
        if self.session.headers.get('Authorization'):
            del self.session.headers['Authorization']

    def test_connection(self) -> Response[str]:
        r = self.session.get('/sapi/v1/health/')
        return r.json()

    def get_24h_price(self, market: str) -> Response[FullDayPricePayload]:
        r = self.session.get('/sapi/v1/market/tickers/',
                             params={market: market})
        return r.json()

    def get_candlestick(self, market: str, limit: Optional[int] = None,
                        period: Optional[int] = None, start_time: Optional[int] = None,
                        end_time: Optional[int] = None) -> Response[CandleStickPayload]:
        loc = locals()
        params = params_to_dict(loc)
        r = self.session.get('/sapi/v1/market/kline/',
                             params=params)
        return r.json()

    def get_orderbook(self, market: str, asks_limit: Optional[int] = None,
                      bids_limit: Optional[int] = None) -> Response[OrderBookPayload]:

        loc = locals()
        params = params_to_dict(loc)
        r = self.session.get('/sapi/v1/market/orderbook/',
                             params=params)
        return r.json()

    def get_recent_trades(self, market: str, limit: Optional[int] = None,
                          timestamp: Optional[int] = None, order_by: Optional[int] = None) -> Response[List[RecentTradesPayload]]:
        loc = locals()
        params = params_to_dict(loc)
        r = self.session.get('/sapi/v1/market/trades/',
                             params=params)
        return r.json()

    def get_nonce(self, eth_address: str) -> Response[str]:
        loc = locals()
        body = params_to_dict(loc)
        r = self.session.post('/sapi/v1/auth/v2/nonce/',
                              json=body)
        return r.json()

    def login(self, eth_address: str, user_signature: str) -> LoginResponse:
        loc = locals()
        body = params_to_dict(loc)
        r = self.session.post('/sapi/v1/auth/v2/login/',
                              json=body)
        js = r.json()
        try:
            self.session.headers.update(
                {'Authorization': f"JWT {js['token']['access']}"})
            self.set_refresh_token(js['token']['refresh'])
        except KeyError:
            raise AuthenticationError('Invalid Credentials')
        return js

    def complete_login(self, eth_address: str, private_key: str) -> LoginResponse:
        try:
            nonce = self.get_nonce(eth_address)
            user_signature = sign_msg(nonce['payload'], private_key)
            login = self.login(eth_address, user_signature)
        except KeyError:
            raise AuthenticationError('Invalid Credentials')
        return login

    def get_profile_info(self) -> Response[ProfileInformationPayload]:
        self.get_auth_status()
        r = self.session.get('/sapi/v1/user/profile/')
        return r.json()

    def get_balance(self, currency: Optional[str] = None) -> Response[Union[Balance, List[Balance]]]:
        self.get_auth_status()
        loc = locals()
        params = params_to_dict(loc)
        r = self.session.get('/sapi/v1/user/balance/', params=params)
        return r.json()

    def get_profit_and_loss(self) -> Response[ProfitAndLossPayload]:
        self.get_auth_status()
        r = self.session.get('/sapi/v1/user/pnl/')
        return r.json()

    def get_auth_status(self) -> bool:
        try:
            if self.session.headers['Authorization']:
                return True
        except KeyError:
            raise AuthenticationError(
                'This is a private endpoint... Please use login() or complete_login() first')
        return False

    def create_order_nonce(self, body: CreateOrderNonceBody) -> Response[CreateOrderNoncePayload]:
        self.get_auth_status()
        r = self.session.post('/sapi/v1/orders/nonce/',
                              json=body)
        return r.json()

    def create_new_order(self, body: CreateNewOrderBody) -> Response[Order]:
        self.get_auth_status()
        r = self.session.post('/sapi/v1/orders/create/',
                              json=body)
        return r.json()

    def create_complete_order(self, nonce: CreateOrderNonceBody, stark_private_key: str):
        nonce_res = self.create_order_nonce(nonce)
        msg_hash = sign_order_with_stark_private_key(stark_private_key, nonce_res['payload'])
        order = self.create_new_order(msg_hash)
        return order

    def get_order(self, order_id: int) -> Response[OrderPayload]:
        self.get_auth_status()
        r = self.session.get(f'/sapi/v1/orders/{order_id}')
        return r.json()

    def list_orders(self, limit: Optional[int] = 50, page: Optional[int] = None, market: Optional[str] = None, ord_type: Optional[str] = None, state: Optional[str] = None, base_unit: Optional[str] = None, quote_unit: Optional[str] = None, start_time: Optional[int] = None, end_time: Optional[int] = None, side: Optional[str] = None) -> Response[List[OrderPayload]]:
        self.get_auth_status()
        loc = locals()
        params = params_to_dict(loc)
        r = self.session.get('/sapi/v1/orders', params=params)
        return r.json()

    def cancel_order(self, order_id: int) -> Response[CancelOrder]:
        self.get_auth_status()
        loc = locals()
        body = params_to_dict(loc)
        r = self.session.post('/sapi/v1/orders/cancel/', json=body)
        return r.json()

    def list_trades(self, limit: Optional[int] = None, page: Optional[int] = None, market: Optional[str] = None, start_time: Optional[int] = None, end_time: Optional[int] = None, order_by: Optional[str] = None) -> Response[List[TradePayload]]:
        self.get_auth_status()
        loc = locals()
        body = params_to_dict(loc)
        r = self.session.get('/sapi/v1/trades/', json=body)
        return r.json()

    def get_coin_stats(self):
        r = self.session.post('/main/stat/v2/coins/')
        return r.json()

    def get_vault_id(self, coin: str):
        self.get_auth_status()
        loc = locals()
        body = params_to_dict(loc)
        r = self.session.post('/main/user/create_vault/', json=body)
        return r.json()

    def list_deposits(self, params: ListDepositParams):
        self.get_auth_status()
        loc = locals()
        body = params_to_dict(loc)
        r = self.session.get('/sapi/v1/deposits/', json=body)
        return r.json()

    def get_token_balance(self, provider: Web3, eth_address: str, currency: str):
        if currency == 'eth':
            balance_wei = provider.eth.get_balance(eth_address) # type: ignore
            balance_eth = provider.from_wei(balance_wei, 'ether')
            return balance_eth

        coin_stats =  self.get_coin_stats()
        current_coin = filter_ethereum_coin(coin_stats['payload'], currency)
        token_contract = current_coin['token_contract']
        decimal = current_coin['decimal']
        
        contract = provider.eth.contract(address=token_contract, abi=Config.ERC20_ABI) # type:ignore
        balance = contract.functions.balanceOf(eth_address).call()
        normal_balance = int(balance) / (10 ** int(decimal)) # type:ignore
        return normal_balance

    def crypto_deposit_start(self, amount: int, stark_asset_id: str, stark_public_key: str, deposit_blockchain_hash: str, deposit_blockchain_nonce: str, vault_id: str):
        amount_to_string = str(amount)
        payload = {
            'amount': amount_to_string,
            'token_id': stark_asset_id,
            'stark_key': stark_public_key,
            'deposit_blockchain_hash': deposit_blockchain_hash,
            'deposit_blockchain_nonce': deposit_blockchain_nonce,
            'vault_id': vault_id
        }
        r = self.session.post('/sapi/v1/payment/stark/start/', json=payload)
        return r.json()

    def approve_unlimited_allowance_ethereum_network(self, coin: str, signer, w3: Web3):
        coin_stats = self.get_coin_stats()
        current_coin = filter_ethereum_coin(coin_stats['payload'], coin)
        token_contract = current_coin['token_contract']
        stark_contract = Config.STARK_CONTRACT[self.option]
        res = approve_unlimited_allowance_util(contract_address=stark_contract, token_contract=token_contract, signer=signer, w3=w3)
        return res


    def deposit_from_ethereum_network_with_stark_key(self, signer: Account, provider: Web3, stark_public_key: str, amount: float, currency: str):
        w3 = provider
        amount = float(amount)
        if amount <= 0:
            raise InvalidAmountError("Please enter a valid amount. It should be a numerical value greater than zero.")

        self.get_auth_status()
        coin_stats = self.get_coin_stats()['payload']
        current_coin = filter_ethereum_coin(coin_stats, currency)
        quanitization = current_coin['quanitization']
        decimal = current_coin['decimal']
        token_contract = current_coin['token_contract']
        stark_asset_id = current_coin['stark_asset_id']

        quantized_amount = amount*(10**int(quanitization))
        quantized_amount = int(quantized_amount)

        vault = self.get_vault_id(currency)
        stark_contract = Config.STARK_CONTRACT[self.option]
        # ABI is what defines the intrinsic properties of a smart contract
        stark_abi = Config.STARK_ABI[self.option]

        contract_instance = w3.eth.contract(address=stark_contract, abi=stark_abi) # type:ignore
        parsed_amount = w3.to_wei(amount, 'ether')
        gwei = w3.to_wei(amount, 'gwei')

        # In the context of building transactions on the Ethereum network, 
        # "overrides" typically refer to parameters that allow you to customize 
        # the behavior of a transaction. These overrides can be used when interacting 
        # with smart contracts, and they provide a way to specify details such as 
        # the amount of gas to use, the gas price, the recipient address, and the value to send.
        overrides = {
            'from': signer.address, # type:ignore
            'nonce': get_nonce(signer, provider)
        }


        balance = self.get_token_balance(provider, signer.address, currency) # type:ignore

        if balance < amount:
            raise BalanceTooLowError(f'Current Balance ({balance}) for "{currency}" is too low, please add balance before deposit')

        stark_public_key_uint = int(get_0x0_to_0x(stark_public_key), 16)
        stark_asset_id_uint = int(get_0x0_to_0x(stark_asset_id), 16)
        if currency == 'eth':
            # On the smart contracts itself (not override parameter)
            # - For transactions involving Ether (ETH), the value is implicit and 
            #     doesn't need to be explicitly specified.
            # - For transactions involving other tokens on the Ethereum network, 
            #     you typically interact with smart contracts, and you need to explicitly specify 
            #     the value and other parameters required by the contract.
            overrides['value'] = w3.to_wei(amount, 'ether')
            transaction_pre_build = contract_instance.functions.depositEth(
                stark_public_key_uint,
                stark_asset_id_uint,
                vault['payload']['id']
            )

        else:
            allowance = get_allowance(user_address=signer.address, stark_contract=stark_contract, token_contract=token_contract, decimal=decimal, w3=provider) # type:ignore
            if allowance < amount:
                raise AllowanceTooLowError(f"Current Allowance ({allowance}) is too low, please use Client.approve_unlimited_allowance_ethereum_network()")
            transaction_pre_build = contract_instance.functions.depositERC20(
                stark_public_key_uint,
                stark_asset_id_uint,
                vault['payload']['id'],
                int(quantized_amount)
            )

        transaction = transaction_pre_build.build_transaction(overrides)
        signed_tx = signer.sign_transaction(transaction)
        # send this signed transaction to blockchain
        w3.eth.send_raw_transaction(signed_tx.rawTransaction).hex()
        deposit_response = signed_tx

        res = self.crypto_deposit_start(
            gwei * 10 if currency == 'eth' else quantized_amount,
            get_0x0_to_0x(stark_asset_id),
            get_0x0_to_0x(stark_public_key),
            deposit_response['hash'].hex(),
            transaction['nonce'],
            vault['payload']['id']
        )

        res['payload'] = {'transaction_hash': deposit_response['hash'].hex()}

        return res

    def start_normal_withdrawal(self, body: dict):
        payload = {
            'amount': body['amount'],
            'token_id': body['symbol']
        }
        r = self.session.post('/sapi/v1/payment/withdrawals/v1/initiate/', json=payload)
        return r.json()

    def validate_normal_withdrawal(self, body: dict):
        r = self.session.post('/sapi/v1/payment/withdrawals/v1/validate/', json=body)
        return r.json()

    def initiate_normal_withdrawal(self, key_pair: dict, amount: float, coin_symbol: str):
        if float(amount)<=0:
            raise InvalidAmountError('Please enter a valid amount. It should be a numerical value greater than zero.')
        self.get_auth_status()
        initiate_response = self.start_normal_withdrawal({'amount': amount, 'symbol': coin_symbol})
        signature = sign_withdrawal_tx_msg_hash(key_pair, initiate_response['payload']['msg_hash'])
        msg_hex = initiate_response['payload']['msg_hash']
        msg_hex = hex(int(msg_hex))

        validate_response = self.validate_normal_withdrawal({
            'msg_hash': str(msg_hex[2:]),
            'signature': signature,
            'nonce': initiate_response['payload']['nonce']
        })
        return validate_response

    def get_pending_normal_withdrawal_amount_by_coin(self, coin_symbol: str, user_public_eth_address: str, signer: Account, provider: Web3):
        self.get_auth_status()
        w3 = provider
        coin_stats = self.get_coin_stats()['payload']
        current_coin = filter_ethereum_coin(coin_stats_payload=coin_stats, coin=coin_symbol)
        stark_asset_id = current_coin['stark_asset_id']
        blockchain_decimal = current_coin['blockchain_decimal']
        stark_contract = Config.STARK_CONTRACT[self.option]
        stark_abi = Config.STARK_ABI[self.option]

        contract_instance = w3.eth.contract(address=stark_contract, abi=stark_abi) # type:ignore
        
        balance = contract_instance.functions.getWithdrawalBalance(
            int(user_public_eth_address, 16),
            int(stark_asset_id, 16)
        ).call()

        return format_withdrawal_amount(amount=balance, decimals=int(blockchain_decimal), symbol=coin_symbol)

    def complete_normal_withdrawal(self, coin_symbol: str, user_public_eth_address: str, signer: Account, provider: Web3, gas_price: int):
        self.get_auth_status()
        w3 = provider
        coin_stats = self.get_coin_stats()['payload']
        current_coin = filter_ethereum_coin(coin_stats_payload=coin_stats, coin=coin_symbol)
        stark_asset_id = current_coin['stark_asset_id']
        stark_contract = Config.STARK_CONTRACT[self.option]
        stark_abi = Config.STARK_ABI[self.option]

        contract_instance = w3.eth.contract(address=stark_contract, abi=stark_abi) # type:ignore

        transaction_pre_build = contract_instance.functions.withdraw(int(user_public_eth_address, 16), int(stark_asset_id, 16))

        gas_price_current = w3.eth.gas_price
        if gas_price_current > gas_price:
            raise InvalidAmountError('Current gas price is higher than provided')

        overrides = {
            'nonce': get_nonce(signer, provider),
            'gasPrice': gas_price,
        }
        transaction = transaction_pre_build.build_transaction(overrides)
        signed_tx = signer.sign_transaction(transaction)
        # send this signed transaction to blockchain
        w3.eth.send_raw_transaction(signed_tx.rawTransaction).hex()
        res = signed_tx
        return res

    def list_normal_withdrawals(self, params: Optional[ListWithdrawalParams]=None):
        self.get_auth_status()
        r = self.session.get('/sapi/v1/payment/withdrawals/', json=params)
        return r.json()

    def initiate_internal_transfer(self, body: InternalTransferInitiateBody):
        r = self.session.post('/sapi/v1/internal_transfers/v2/initiate/', json=body)
        return r.json()

    def execute_internal_transfers(self, body: InternalTransferProcessBody):
        r = self.session.post('/sapi/v1/internal_transfers/v2/process/', json=body)
        return r.json()

    def intiate_and_process_internal_transfers(self, key_pair: dict, organization_key: str, api_key: str, currency: str, amount: float,
                                            destination_address: str, client_reference_id: Optional[int]=None):
        self.get_auth_status()
        loc = locals()
        body = params_to_dict(loc)
        initiate_response = self.initiate_internal_transfer(body) # type: ignore
        signature = sign_internal_tx_msg_hash(key_pair, initiate_response['payload']['msg_hash'])
        execute_response = self.execute_internal_transfers({
            'organization_key': organization_key,
            'api_key': api_key,
            'signature': signature,
            'nonce': int(initiate_response['payload']['nonce']),
            'msg_hash': initiate_response['payload']['msg_hash']
        })
        return execute_response

    def list_internal_transfers(self, params: Optional[ListInternalTransferParams]=None):
        self.get_auth_status()
        r = self.session.get('/sapi/v1/internal_transfers/v2/', params=params)  # type:ignore
        return r.json()

    def get_internal_transfer_by_client_id(self, client_reference_id: int):
        self.get_auth_status()
        r = self.session.get(f'/sapi/v1/internal_transfers/v2/{client_reference_id}')
        return r.json()

    def check_internal_transfer_user_exists(self, organization_key: str, api_key: str, destination_address: str):
        self.get_auth_status()
        loc = locals()
        body = params_to_dict(loc)
        r = self.session.post('/sapi/v1/internal_transfers/v2/check_user_exists/', json=body)
        return r.json()

    def get_network_config(self):
        r = self.session.post('/main/stat/v2/app-and-markets/')
        return r.json()['payload']['network_config']

    def get_polygon_token_balance(self, provider: Web3, eth_address: str, currency: str):
        if currency == 'matic':
            balance = provider.eth.get_balance(eth_address)     # type:ignore
            return Web3.from_wei(balance, 'ether')
        network_config = self.get_network_config()
        polygon_config = network_config['POLYGON']
        allowed_tokens = polygon_config['tokens']

        current_coin = filter_cross_chain_coin(polygon_config, currency, 'TOKENS')

        decimal = current_coin['blockchain_decimal']
        token_contract = current_coin['token_contract']
        contract = provider.eth.contract(address=token_contract, abi=Config.ERC20_ABI)
        
        balance = contract.functions.balanceOf(eth_address).call()
        normal_balance = balance / (10 ** int(decimal))
        return normal_balance

    def cross_chain_deposit_start(self, amount: float, currency: str, deposit_blockchain_hash: str, deposit_blockchain_nonce: str):
        amount_to_string = str(amount)
        body = {
            'amount': amount_to_string,
            'currency': currency,
            'network': 'POLYGON',
            'deposit_blockchain_hash': deposit_blockchain_hash,
            'deposit_blockchain_nonce': deposit_blockchain_nonce,
        }
        r = self.session.post('/sapi/v1/deposits/crosschain/create/', json=body)
        return r.json()

    def approve_unlimited_allowance_polygon_network(self, coin: str, signer: Account, w3: Web3):
        network_config = self.get_network_config()
        polygon_config = network_config['POLYGON']
        allowed_tokens = polygon_config['tokens']
        contract_address = polygon_config['deposit_contract']

        current_coin = filter_cross_chain_coin(polygon_config, coin, 'DEPOSIT')
        token_contract = current_coin['token_contract']
        res = approve_unlimited_allowance_util(contract_address=contract_address, token_contract=token_contract, signer=signer, w3=w3)
        return res

    def deposit_from_polygon_network_with_signer(self, signer: Account, provider: Web3, currency:str, amount: float):
        if float(amount)<=0:
            raise InvalidAmountError('Please enter a valid amount. It should be a numerical value greater than zero.')

        self.get_auth_status()

        w3 = provider

        network_config = self.get_network_config()
        polygon_config = network_config['POLYGON']
        allowed_tokens = polygon_config['tokens']
        contract_address = polygon_config['deposit_contract']

        current_coin = filter_cross_chain_coin(polygon_config, currency, 'DEPOSIT')

        decimal = current_coin.get('blockchain_decimal')
        token_contract = current_coin.get('token_contract')

        quantized_amount = amount*(10**int(decimal))
        quantized_amount = int(quantized_amount)

        polygon_contract = w3.eth.contract(address=contract_address, abi=Config.POLYGON_ABI['abi'])

        parsed_amount = w3.to_wei(amount, 'ether')
        gwei = w3.from_wei(parsed_amount, 'gwei')
        nonce = get_nonce(signer=signer, provider=provider)

        params = {
            'from': signer.address,     # type:ignore
            'nonce': nonce
        }

        balance = self.get_polygon_token_balance(provider, signer.address, currency)    # type:ignore

        if balance<amount:
            raise BalanceTooLowError(f'Current Balance {balance} for {currency} is too low, please add balance before deposit')

        deposit_response = None

        if currency == 'matic':
            params['value']=parsed_amount
            transaction_pre_build = polygon_contract.functions.depositNative()

        else:
            allowance = get_allowance(user_address=signer.address, stark_contract=contract_address, token_contract=token_contract, decimal=decimal, w3=provider)    # type:ignore
            if allowance < amount:
                raise AllowanceTooLowError(f"Current Allowance ({allowance}) is too low, please use Client.approve_unlimited_allowance_polygon_network")
            transaction_pre_build = polygon_contract.functions.deposit(
                token_contract,
                int(quantized_amount)
            )

        transaction = transaction_pre_build.build_transaction(params)    # type:ignore
        signed_tx = signer.sign_transaction(transaction)
        # send this signed transaction to blockchain
        w3.eth.send_raw_transaction(signed_tx.rawTransaction).hex()
        deposit_response = signed_tx
        res = self.cross_chain_deposit_start(
            amount,
            currency,
            deposit_response['hash'].hex(),
            transaction['nonce']    # type:ignore
        )

        res['payload'] = {'transaction_hash': deposit_response['hash'].hex()}
        return res
    
    def deposit_from_polygon_network(self, rpc_url: str, eth_private_key: str, currency: str, amount: float):
        self.get_auth_status()
        provider = Web3(Web3.HTTPProvider(rpc_url))
        provider.middleware_onion.inject(geth_poa_middleware, layer=0)
        signer = Account.from_key(eth_private_key)
        return self.deposit_from_polygon_network_with_signer(signer=signer, provider=provider, currency=currency, amount=amount)

    def start_fast_withdrawal(self, body: InitiateWithdrawalPayload):
        r = self.session.post('/sapi/v1/payment/fast-withdrawals/v2/initiate/',
                                json={'amount': body['amount'], 'token_id': body['symbol'], 'network': body['network']})
        return r.json()

    def process_fast_withdrawal(self, body: ProcessFastWithdrawalPayload):
        r = self.session.post('/sapi/v1/payment/fast-withdrawals/v2/process/', json=body)
        return r.json()

    def fast_withdrawal(self, key_pair: dict, amount: float, coin_symbol: str, network: str):
        if amount<=0:
            raise InvalidAmountError('Please enter a valid amount. It should be a numerical value greater than zero.')
        self.get_auth_status()
        if network == 'POLYGON':
            network_config = self.get_network_config()
            polygon_config = network_config['POLYGON']
            _ = filter_cross_chain_coin(polygon_config, coin_symbol, 'WITHDRAWAL')
        else:
            coin_stats = self.get_coin_stats()['payload']
            _ = filter_ethereum_coin(coin_stats, coin_symbol)
        
        initiate_response = self.start_fast_withdrawal({
            'amount': amount,
            'symbol': coin_symbol,
            'network': network
        })
        signature = sign_withdrawal_tx_msg_hash(key_pair, str(int(initiate_response['payload']['msg_hash'], 16)))

        validate_response = self.process_fast_withdrawal({
            'msg_hash': initiate_response['payload']['msg_hash'],
            'signature': signature,     # type:ignore
            'fastwithdrawal_withdrawal_id': initiate_response['payload']['fastwithdrawal_withdrawal_id'],
        })
        return validate_response

    def list_fast_withdrawals(self, params: Optional[ListWithdrawalParams]=None):
        self.get_auth_status()
        r = self.session.get('/sapi/v1/payment/fast-withdrawals/', params=params)   # type:ignore
        return r.json()



import responses
import sys
import os
import pytest
import requests
from typing import List, cast
from responses import matchers
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from src.tanxconnector import Client  # noqa: E402
from src.tanxconnector import sign_order_with_stark_private_key  # noqa: E402
from src.tanxconnector.typings import Balance  # noqa: E402
from src.tanxconnector.exception import InvalidAmountError, BalanceTooLowError
from tests.mock_responses import *
from web3 import EthereumTesterProvider, Account, Web3
BASE_URL = 'https://api.tanx.fi'

# load_dotenv()
PRIVATE_KEY = '7d6384d6877be027aa25bd458f2058e3f7ff68347dc583a9baf96f5f97b413a8'
ETH_ADDRESS = '0x713Cf80b7c71440E7a09Dede1ee23dCBf862fB66'

client = Client()
client2 = Client()


@responses.activate
def test_ping():
    responses.get(f'{BASE_URL}/sapi/v1/health/', json={
        "status": 'success',
        "message": 'Working fine!',
        "payload": '',
    })
    assert "Working" in client.test_connection()['message']


@responses.activate
def test_get_24h_price():
    responses.get(f'{BASE_URL}/sapi/v1/market/tickers/',
                  json={
                      "status": 'success',
                      "message": 'Retrieval Successful',
                      "payload": {
                          "btcusdc": {"at": '1681903275', "ticker": [{}]},
                          "btcusdt": {"at": '1681903275', "ticker": [{}]},
                          "ethusdc": {"at": '1681903275', "ticker": [{}]},
                          "ethusdt": {"at": '1681903275', "ticker": [{}]},
                          "usdcusdt": {"at": '1681903275', "ticker": [{}]},
                      },
                  }
                  )
    assert "btcusdc" in client.get_24h_price('btcusdc')['payload']


@responses.activate
def test_get_candlestick():
    responses.get(f'{BASE_URL}/sapi/v1/market/kline/', json={
        "status": 'success',
        "message": 'Retrieval Successful',
        "payload": [
            [1681689600, 2014.02, 2014.02, 2014.02, 2014.02, 0],
            [1681696800, 2014.02, 2014.02, 2014.02, 2014.02, 0],
            [1681704000, 2014.02, 2014.02, 2014.02, 2014.02, 0],
            [1681711200, 2014.02, 2014.02, 2014.02, 2014.02, 0.01],
            [1681718400, 2014.02, 2014.02, 2014.02, 2014.02, 0],
            [1681725600, 2014.02, 2014.02, 2014.02, 2014.02, 0],
            [1681732800, 2014.02, 2014.02, 2014.02, 2014.02, 0],
            [1681740000, 2014.02, 2014.02, 2014.02, 2014.02, 0],
            [1681747200, 2014.02, 2014.02, 2014.02, 2014.02, 0.1618],
            [1681754400, 2014.02, 2014.02, 2014.02, 2014.02, 0],
            [1681761600, 2014.02, 2014.02, 2014.02, 2014.02, 0.008400000000000001],
            [1681768800, 2014.02, 2014.02, 2014.02, 2014.02, 0],
            [1681776000, 2014.02, 2014.02, 2014.02, 2014.02, 0],
            [1681783200, 2014.02, 2014.02, 2014.02, 2014.02, 0],
            [1681790400, 2014.02, 2014.02, 2014.02, 2014.02, 0.1],
            [1681797600, 2014.02, 2014.02, 2014.02, 2014.02, 0.3694],
            [1681804800, 2014.02, 2014.02, 2014.02, 2014.02, 0.05],
            [1681812000, 2014.02, 2014.02, 2014.02, 2014.02, 0.0993],
            [1681819200, 2014.02, 2014.02, 2014.02, 2014.02, 0.015],
            [1681826400, 2014.02, 2014.02, 2014.02, 2014.02, 0.0066],
            [1681833600, 2014.02, 2014.02, 2014.02, 2014.02, 0],
            [1681840800, 2014.02, 2014.02, 2014.02, 2014.02, 0],
            [1681848000, 2014.02, 2014.02, 2014.02, 2014.02, 0.0005],
            [1681855200, 2014.02, 2014.02, 2014.02, 2014.02, 0],
            [1681862400, 2014.02, 2014.02, 2014.02, 2014.02, 0],
            [1681869600, 2014.02, 2014.02, 2014.02, 2014.02, 0],
            [1681876800, 2014.02, 2014.02, 2014.02, 2014.02, 0],
            [1681884000, 2014.02, 2014.02, 2014.02, 2014.02, 0.01],
            [1681891200, 2014.02, 2014.02, 2014.02, 2014.02, 0.0993],
            [1681898400, 2014.02, 2014.02, 2014.02, 2014.02, 0],
        ],
    })
    assert "Retrieval" in client.get_candlestick('btcusdc')['message']


@responses.activate
def test_get_candlestick_raises_400_error():
    responses.get(f'{BASE_URL}/sapi/v1/market/kline/', json={
        "status": 'error',
        "message": 'please enter a valid market',
        "payload": '',
    }, status=400)
    with pytest.raises(requests.exceptions.HTTPError):
        client.get_candlestick('test')['message']


@responses.activate
def test_get_orderbook():
    responses.get(f'{BASE_URL}/sapi/v1/market/orderbook/', json={
        "status": 'success',
        "message": 'Retrieval Successful',
        "payload": {
            "asks": [
                {
                    "id": 100,
                    "uuid": '6f11fcb2-1ef5-4fe3-852d-c73ddc28a8d1',
                    "side": 'sell',
                    "ord_type": 'limit',
                    "price": '2000.0',
                    "avg_price": '0.0',
                    "state": 'wait',
                    "market": 'ethusdc',
                    "created_at": '2023-04-19T11:02:50+02:00',
                    "updated_at": '2023-04-19T11:02:51+02:00',
                    "origin_volume": '0.001',
                    "remaining_volume": '0.001',
                    "executed_volume": '0.0',
                    "maker_fee": '0.001',
                    "taker_fee": '0.001',
                    "trades_count": 0,
                }
            ],
            "bids": [
                {
                    "id": 94,
                    "uuid": '06755d51-fcee-4936-b847-2d4da33588ba',
                    "side": 'buy',
                    "ord_type": 'limit',
                    "price": '1990.0',
                    "avg_price": '0.0',
                    "state": 'wait',
                    "market": 'ethusdc',
                    "created_at": '2023-04-19T11:01:18+02:00',
                    "updated_at": '2023-04-19T11:01:18+02:00',
                    "origin_volume": '0.001',
                    "remaining_volume": '0.001',
                    "executed_volume": '0.0',
                    "maker_fee": '0.001',
                    "taker_fee": '0.001',
                    "trades_count": 0,
                },
            ],
        },
    })
    payload = client.get_orderbook('btcusdc')['payload']
    assert "asks" in payload
    assert "bids" in payload


def test_get_orderbook_raises_type_error():
    with pytest.raises(TypeError):
        client.get_orderbook()['payload'] # type:ignore


@responses.activate
def test_get_recent_trades():
    responses.get(f'{BASE_URL}/sapi/v1/market/trades/', json={
        "status": 'success',
        "message": 'Retrieval Successful',
        "payload": [
            {
                "id": 40,
                "price": 1990,
                "amount": 0.001,
                "total": 1.99,
                "market": 'ethusdc',
                "created_at": 1681894870,
                "taker_type": 'buy',
            }
        ]})
    assert "amount" in client.get_recent_trades('btcusdc')['payload'][0]


@responses.activate
def test_complete_login():
    responses.post(url=f"{BASE_URL}/sapi/v1/auth/v2/nonce/",
                   json={'status': 'success', 'message': 'Cached Nonce Acquired',
                         'payload': 'You’re now signing into Tanx Testnet, make sure the origin is https://testnet.tanx.fi (Login-code:abc)'},
                   status=200)
    responses.post(url=f"{BASE_URL}/sapi/v1/auth/v2/login/",
                   json={'status': 'success', 'message': 'Login Successful', 'payload': {
                       'uid': ''}, 'token': {'refresh': 'test', 'access': 'test'}},
                   status=200)
    assert "token" in client.complete_login(ETH_ADDRESS, PRIVATE_KEY)
    assert "token" in client2.complete_login(ETH_ADDRESS, PRIVATE_KEY)


@responses.activate
def test_complete_login_raises_invalid_eth_address_400_error():
    responses.post(url=f"{BASE_URL}/sapi/v1/auth/v2/nonce/",
                   json={
                       'status': 'error', 'message': 'Ensure eth_address has at least 30 characters.', 'payload': ''},
                   status=400)
    with pytest.raises(requests.exceptions.HTTPError):
        print(client.complete_login('test', PRIVATE_KEY))


@responses.activate
def test_complete_login_raises_incorrect_eth_address_400_error():
    responses.post(url=f"{BASE_URL}/sapi/v1/auth/v2/nonce/",
                   json={'status': 'success', 'message': 'Cached Nonce Acquired',
                         'payload': 'You’re now signing into Tanx Testnet, make sure the origin is https://testnet.tanx.fi (Login-code:abc)'},
                   status=200)
    responses.post(url=f"{BASE_URL}/sapi/v1/auth/v2/login/",
                   json={
                       'status': 'error', 'message': 'Invalid Credentials or Token Expired, Kindly login again.', 'payload': ''},
                   status=400)
    with pytest.raises(requests.exceptions.HTTPError):
        client.complete_login(
            '0x83D37295507F7C838f6a7dd1E41a4A81aC7C9a5E', PRIVATE_KEY)


@responses.activate
def test_get_profile_info():
    responses.get(url=f"{BASE_URL}/sapi/v1/user/profile/",
                  json={'status': 'success', 'message': 'Successful', 'payload': {
                      'name': 'test', 'img': None, 'username': '0x7tanx', 'stark_key': '0x'}},
                  status=200)
    assert "name" in client.get_profile_info()['payload']


@responses.activate
def test_get_balance():
    responses.get(url=f"{BASE_URL}/sapi/v1/user/balance/",
                  json={'status': 'success', 'message': 'Retrieved Successfully', 'payload': [{'currency': 'btc', 'balance': '0.0091892', 'locked': '0.0'}, {
                      'currency': 'eth', 'balance': '0.2179175', 'locked': '0.0'}, {'currency': 'usdc', 'balance': '56.354237624', 'locked': '0.0'}, {'currency': 'usdt', 'balance': '79.699671653', 'locked': '0.0'}]},
                  status=200)
    r = cast(List[Balance], client.get_balance()['payload'])
    assert "balance" in r[0]


@responses.activate
def test_get_balance_expired_access_token():

    responses.get(url=f"{BASE_URL}/sapi/v1/user/balance/",
                  json={
                      "status": 'error',
                      "message": 'Given token not valid for any token type',
                      "payload": {
                          "token_class": 'AccessToken',
                          "token_type": 'access',
                          "message": 'Token is invalid or expired',
                          "code": 'token_not_valid',
                      },
                  },
                  status=401)

    responses.post(url=f"{BASE_URL}/sapi/v1/auth/token/refresh/", 
                   json={
                        "status": 'success',
                        "message": '',
                        "payload": {
                            "access": 'ferasdfklre',
                            "refresh": 'fekcjicbd',
                        },
                    },
                    status=200)
                
    responses.get(url=f"{BASE_URL}/sapi/v1/user/balance/",
                json={'status': 'success', 'message': 'Retrieved Successfully', 'payload': [{'currency': 'btc', 'balance': '0.0091892', 'locked': '0.0'}, {
                    'currency': 'eth', 'balance': '0.2179175', 'locked': '0.0'}, {'currency': 'usdc', 'balance': '56.354237624', 'locked': '0.0'}, {'currency': 'usdt', 'balance': '79.699671653', 'locked': '0.0'}]},
                status=200)
    
    r = cast(List[Balance], client2.get_balance()['payload'])

    assert "balance" in r[0]

@responses.activate
def test_get_balance_expired_refresh_token():
    responses.get(url=f"{BASE_URL}/sapi/v1/user/balance/",
                  json={
                      "status": 'error',
                      "message": 'Given token not valid for any token type',
                      "payload": {
                          "token_class": 'AccessToken',
                          "token_type": 'access',
                          "message": 'Token is invalid or expired',
                          "code": 'token_not_valid',
                      },
                  },
                 status=401)

    responses.post(url=f"{BASE_URL}/sapi/v1/auth/token/refresh/", json={
        "status": "error",
        "message": "Token is invalid or expired",
        "payload": {
            "code": "token_not_valid"
        }
    },
      status=401)
    

    with pytest.raises(requests.exceptions.HTTPError):
        r = client2.get_balance()['payload']


@responses.activate
def test_get_profit_and_loss():
    responses.get(url=f"{BASE_URL}/sapi/v1/user/pnl/",
                  json={'status': 'success',
                        'message': 'Retrieval Successful', 'payload': [{}]},
                  status=200)
    assert "Retrieval" in client.get_profit_and_loss()[
        'message']


def test_get_auth_status():
    assert client.get_auth_status()


@responses.activate
def test_create_complete_order():
    responses.post(url=f"{BASE_URL}/sapi/v1/orders/nonce/",
                   json={'status': 'success', 'message': 'order nonce created successfully, please sign the msg_hash to place the order',
                         'payload': {'nonce': 35179980, 'msg_hash': '0xe5eea5423fbb8332c24d750bc49c899eba99af665306aaddc9968972002e1'}},
                   status=200)
    responses.post(url=f"{BASE_URL}/sapi/v1/orders/create/",
                   json={'status': 'success', 'message': 'Created Order Successfully',
                         'payload': {}},
                   status=200)
    nonce_res = client.create_order_nonce({'market': 'btcusdt', 'ord_type': 'market',
                                           'price': 29580.51, 'side': 'buy', 'volume': 0.0001})
    stark_private_key = '0x64004f706c1eaa39348afb3191c74812d86e5b14b967e578addb4d89ce1234c'
    # replace with your stark private key
    msg_hash = sign_order_with_stark_private_key(
        stark_private_key, nonce_res['payload'])
    assert "Order" in client.create_new_order(msg_hash)['message']


@responses.activate
def test_create_order_nonce_raises_400_error():
    responses.post(url=f"{BASE_URL}/sapi/v1/orders/nonce/",
                   json={
                       'status': 'error', 'message': 'Maximum decimals allowed for volume is 4 in btcusdt market', 'payload': ''},
                   status=400)
    with pytest.raises(requests.exceptions.HTTPError):
        nonce_res = client.create_order_nonce({'market': 'btcusdt', 'ord_type': 'market',
                                               'price': 29580.51, 'side': 'buy', 'volume': 0.00001})
        stark_private_key = '0x64004f706c1eaa39348afb3191c74812d86e5b14b967e578addb4d89ce1234c'
        # replace with your stark private key
        msg_hash = sign_order_with_stark_private_key(
            stark_private_key, nonce_res['payload'])
        client.create_new_order(msg_hash)['message']


@responses.activate
def test_get_order():
    responses.get(url=f"{BASE_URL}/sapi/v1/orders/1234",
                  json={'status': 'success',
                        'message': 'Order Retrieved Successfully', 'payload': {}},
                  status=200)

    assert "Order" in client.get_order(1234)['message']


@responses.activate
def test_list_orders():
    responses.get(url=f"{BASE_URL}/sapi/v1/orders",
                  json={'status': 'success',
                        'message': 'Orders Retrieved Successfully', 'payload': {}},
                  status=200)

    assert "Orders" in client.list_orders()['message']


@responses.activate
def test_cancel_order():
    responses.post(url=f"{BASE_URL}/sapi/v1/orders/cancel/",
                   json={'status': 'success',
                         'message': 'Order is successfully queued for cancellation', 'payload': {}},
                   status=200)
    assert "Order" in client.cancel_order(1234)['message']


@responses.activate
def test_list_trades():
    responses.get(url=f"{BASE_URL}/sapi/v1/trades/",
                  json={'status': 'success',
                        'message': 'Trades Retrieved Successfully', 'payload': [{}]},
                  status=200)
    assert "Trades" in client.list_trades()['message']

@responses.activate
def test_eth_deposits_with_stark_key_success():
    responses.post(url=f'{BASE_URL}/sapi/v1/payment/stark/start/',
                    json={'status': 'success',
                        'message': 'Success! Awaiting Blockchain Confirmation',
                        'payload': ''})
    res = client.crypto_deposit_start(100000,'0x27..','0x27..','0x67..','930','65707',)
    assert 'status' in res
    assert 'success' == res['status']
    assert 'payload' in res

@responses.activate
def test_eth_deposits_with_stark_key_fail():
    responses.post(url=f'{BASE_URL}/main/payment/stark/start/',
                    json={'status':'error',
                        'message':'Essential parameters are missing',
                        'payload': ''})
    with pytest.raises(requests.exceptions.RequestException):
        res = client.crypto_deposit_start(100000,'0x27..','0x27..','0x67..','930','65707',)
    # Handle exception
        
        data = res.response.json()
        assert 'status' in data
        assert data['status'] == 'error'
        assert 'Essential parameters' in data['message']

@responses.activate
def test_list_deposits():
    responses.get(url=f'{BASE_URL}/sapi/v1/deposits/',
                    json=list_deposits_response)
    data = client.list_deposits({'network': 'ETHEREUM'}) # type:ignore

    assert 'status' in data
    assert data['status'] == 'success'
    assert 'payload' in data

@responses.activate
def test_initiate_withdrawals_success():
    responses.post(url=f'{BASE_URL}/sapi/v1/payment/withdrawals/v1/initiate/',
                    json={'status': 'success',
                        'message': 'successfully initiated withdrawal',
                        'payload': {
                            'nonce': 7819,
                            'msg_hash': '686148137588728084357640508492604406021032862346002124816784805415214096923'
                        }})
    res = client.start_normal_withdrawal({
        'amount': 0.00001,
        'symbol': 'eth'
    })
    assert 'status' in res
    assert res['status'] == 'success'
    assert 'payload' in res

@responses.activate
def test_validate_withdrawal_success():
    responses.post(url=f'{BASE_URL}/sapi/v1/payment/withdrawals/v1/validate/',
                    json=validate_withdrawal_response)
    res = client.validate_normal_withdrawal({
            'msg_hash':
                '1845898ec19c65beac9eb12be93adc8fa4fe00a494aa005e2f2cc5bade3a21b',
            'signature': {
                'r': '0x4a0a8a...',
                's': '0x7fc5d01...',
                'recoveryParam': 1,
            },
            'nonce': '7819',
        })
    
    assert 'status' in res
    assert res['status'] == 'success'
    assert 'payload' in res

@responses.activate
def test_validate_withdrawal_failure():
    responses.post(url=f'{BASE_URL}/sapi/v1/payment/withdrawals/v1/validate/',
                    json={'status': 'error',
                    'message': 'Withdrawal Validation Failed, please try again',
                    'payload': ''})
    res = client.validate_normal_withdrawal({
            'msg_hash':
                '1845898ec19c65beac9eb12be93adc8fa4fe00a494aa005e2f2cc5bade3a21b',
            'signature': {
                'r': '0x4a0a8a...',
                's': '0x7fc5d01...',
                'recoveryParam': 1,
            },
            'nonce': '7819',
        })
    
    assert 'status' in res
    assert res['status'] == 'error'
    assert 'Withdrawal Validation Failed' in res['message']

@responses.activate
def test_list_withdrawals():
    responses.get(url=f'{BASE_URL}/sapi/v1/payment/withdrawals/',
                    json=list_withdrawals_response)
    res = client.list_normal_withdrawals()

    assert 'status' in res
    assert res['status'] == 'success'
    assert 'payload' in res
def test_deposit_from_ethereum_network_with_stark_key_invalid_amount():
    w3 = Web3()
    test_signer = w3.eth.account.create()

    test_provider = Web3(EthereumTesterProvider())

    with pytest.raises(InvalidAmountError):
        client.deposit_from_ethereum_network_with_stark_key(signer=test_signer, provider=test_provider, stark_public_key='0x27..', amount=0, currency='eth')

@responses.activate
def test_deposit_from_ethereum_network_with_stark_key_low_balance():
    responses.post(f'{BASE_URL}/main/stat/v2/coins/', json=coin_stats_response)
    responses.post(f'{BASE_URL}/main/user/create_vault/', json=get_vault_id_response)

    test_provider = Web3(EthereumTesterProvider())
    w3 = Web3()
    test_signer = w3.eth.account.create()

    with pytest.raises(BalanceTooLowError):
        client.deposit_from_ethereum_network_with_stark_key(signer=test_signer, provider=test_provider, stark_public_key='0x27..', amount=0.0001, currency='eth')

@responses.activate
def test_initiat_internal_transfer_succcess():
    responses.post(url=f'{BASE_URL}/sapi/v1/internal_transfers/v2/initiate/',
                    json={'status': 'success',
                        'message': 'Please sign the message, to complete the transaction',
                        'payload': {
                            'nonce': 797982128,
                            'msg_hash': '0x2af4af1aa1e47a8b4d71a111a0b5a0649d80d586090548f7bb5a7ba74c287d3',
                        }
                    })
    res = client.initiate_internal_transfer({
        'organization_key': 'abcd...',
        'api_key': 'xyz...',
        'currency': 'usdc',
        'amount': 1,
        'destination_address': '0xF5...',
        'client_reference_id': 10
    })

    assert 'status' in res
    assert res['status'] == 'success'
    assert 'payload' in res
    assert 'nonce' in res['payload']

@responses.activate
def test_execute_internal_transfer_success():
    responses.post(url=f'{BASE_URL}/sapi/v1/internal_transfers/v2/process/',
                    json={
                        'status': 'success',
                        'message': 'Internal transfer processed successfully',
                        'payload': {
                            'client_reference_id': '6462569061361987',
                            'amount': '1.0000000000000000',
                            'currency': 'usdc',
                            'from_address': '0x6c875514E42F14B891399A6a8438E6AA8F77B178',
                            'destination_address': '0xF5F467c3D86760A4Ff6262880727E854428a4996',
                            'status': 'success',
                            'created_at': '2023-07-26T06:28:52.350343Z',
                            'updated_at': '2023-07-26T06:28:52.902831Z'
                        }
                    })
    res = client.execute_internal_transfers({
        'organization_key': 'abcd...',
        'api_key': 'xyz...',
        'signature': {
            'r': '0x12...',
            's': '0x4a...',
        },
        'nonce': 14214931,
        'msg_hash': '0xda0...'
    }) # type:ignore
    assert 'status' in res
    assert res['status'] == 'success'
    assert 'payload' in res
    assert 'client_reference_id' in res['payload']    

@responses.activate
def test_initiate_internal_transfer_403_invalid_credential():
    responses.post(f'{BASE_URL}/sapi/v1/internal_transfers/v2/initiate/',
                    json={
                        'status': 'error',
                        'message': 'Invalid organization key',
                        'payload': ''
                    }, status=403)
    with pytest.raises(requests.exceptions.HTTPError):
        res = client.initiate_internal_transfer({
            'organization_key': 'abcd...',
            'api_key': 'xyz...',
            'currency': 'usdc',
            'amount': 1,
            'destination_address': '0xF5...',
            'client_reference_id': 10
        })
        assert 'status' in res
        assert res['status'] == 'error'

@responses.activate
def test_execute_internal_transfer_403_invalid_credential():
    responses.post(f'{BASE_URL}/sapi/v1/internal_transfers/v2/process/',
                    json={
                        'status': 'error',
                        'message': 'Invalid organization key',
                        'payload': ''
                    }, status=403)
    
    with pytest.raises(requests.exceptions.HTTPError):
        res = client.execute_internal_transfers({
            'organization_key': 'abcd...',
            'api_key': 'xyz...',
            'signature': {
                'r': '0x12...',
                's': '0x4a...',
            },
            'nonce': 14214931,
            'msg_hash': '0xda0...'
        }) # type:ignore
        assert 'status' in res
        assert res['status'] == 'error'

@responses.activate
def test_list_internal_transfers_success():
    responses.get(url=f'{BASE_URL}/sapi/v1/internal_transfers/v2/',
                    json=list_internal_transfer_response, status=200)
    res = client.list_internal_transfers()
    assert 'status' in res
    assert res['status']=='success'
    assert 'client_reference_id' in res['payload']['internal_transfers'][0]

@responses.activate
def test_get_internal_transfer_by_client_id():
    client_id = 1234
    responses.get(url=f'{BASE_URL}/sapi/v1/internal_transfers/v2/{client_id}',
                    json=get_internal_transfer_by_client_id_response, status=200)
    res = client.get_internal_transfer_by_client_id(client_reference_id=client_id)
    assert 'status' in res
    assert res['status'] == 'success'

@responses.activate
def test_transfer_does_not_exist():
    client_id = 1234
    responses.get(url=f'{BASE_URL}/sapi/v1/internal_transfers/v2/{client_id}',
                    json={
                        'status': 'error',
                        'message': 'Transfer does not exist',
                        'payload': ''
                    }, status=404)
    
    with pytest.raises(requests.exceptions.HTTPError):
        res = client.get_internal_transfer_by_client_id(client_reference_id=client_id)
        assert 'status' in res
        assert res['status'] == 'error'

@responses.activate
def test_user_exists_success():
    responses.post(url=f'{BASE_URL}/sapi/v1/internal_transfers/v2/check_user_exists/',
                    json={
                        'status': 'success',
                        'message': 'User exists',
                        'payload': {
                            'destination_address': '0xF5...',
                            'exists': True
                        }
                    })

    res = client.check_internal_transfer_user_exists(
        organization_key='abcd...',
        api_key='xyz...',
        destination_address='0xF5...'
    )
    assert 'status' in res
    assert res['status'] == 'success'
    assert 'payload' in res

@responses.activate
def test_user_exists_failure():
    responses.post(url=f'{BASE_URL}/sapi/v1/internal_transfers/v2/check_user_exists/',
                    json={
                        'status': 'success',
                        'message': 'User does not exists',
                        'payload': {
                            'destination_address': '0xF5...',
                            'exists': False
                        }
                    }, status=404)

    with pytest.raises(requests.exceptions.HTTPError):
        res = client.check_internal_transfer_user_exists(
            organization_key='abcd...',
            api_key='xyz...',
            destination_address='0xF5...'
        )
        assert 'status' in res
        assert res['status'] == 'error'

@responses.activate
def test_start_polygon_deposits_success():
    responses.post(url=f'{BASE_URL}/sapi/v1/deposits/crosschain/create/',
                    json={'status': 'success',
                        'message': 'Success! Awaiting Blockchain Confirmation',
                        'payload': {
                            'transaction_hash': ''
                        },
                    })

    res = client.cross_chain_deposit_start(100000,'0x27..','0x67..','930',)

    assert 'status' in res
    assert res['status'] == 'success'
    assert 'payload' in res

@responses.activate
def test_start_polygon_deposits_failure():
    responses.post(url=f'{BASE_URL}/sapi/v1/deposits/crosschain/create/',
                    json={'status': 'error',
                        'message': 'Essential parameters are missing',
                        'payload': '',
                    })
    
    data = client.cross_chain_deposit_start(100000,'0x27..','0x67..','930',)
    assert 'status' in data
    assert data['status'] == 'error'
    assert 'Essential parameters' in data['message']

@responses.activate
def test_list_polygon_deposits():
    responses.get(url=f'{BASE_URL}/sapi/v1/deposits/',
                    json=list_polygon_deposits_response)

    res = client.list_deposits({'network': 'POLYGON'})      # type:ignore
    assert 'status' in res
    assert res['status'] == 'success'
    assert 'payload' in res

@responses.activate
def test_deposit_from_polygon_network_with_signer_invalid_amount():
    w3 = Web3()
    test_signer = w3.eth.account.create()

    test_provider = Web3(EthereumTesterProvider())

    with pytest.raises(InvalidAmountError):
        client.deposit_from_polygon_network_with_signer(signer=test_signer, provider=test_provider, amount=0, currency='matic')

@responses.activate
def test_deposit_from_polygon_network_with_signer_low_balance():
    responses.post(f'{BASE_URL}/main/stat/v2/app-and-markets/', json=network_config_response)
    responses.post(f'{BASE_URL}/main/user/create_vault/', json=get_vault_id_response)

    test_provider = Web3(EthereumTesterProvider())
    w3 = Web3()
    test_signer = w3.eth.account.create()

    with pytest.raises(BalanceTooLowError):
        client.deposit_from_polygon_network_with_signer(signer=test_signer, provider=test_provider, amount=0.0001, currency='matic')




import json
import os

path_ = os.path.dirname(__file__)

with open(f'{path_}/bin/starkex_abi_main.json') as f:
    starkex_abi_main = json.load(f)

with open(f'{path_}/bin/starkex_abi_test.json') as f:
    starkex_abi_test = json.load(f)

with open(f'{path_}/bin/erc20_abi.json') as f:
    erc20Abi = json.load(f)

with open(f'{path_}/bin/polygon_deposit.json') as f:
    polygon_deposit_abi = json.load(f)


class Config:
    ALLOWED_NETWORK_NAME_MAINNET = "mainnet"
    ALLOWED_NETWORK_NAME_SEPOLIA = "sepolia"
    STARK_CONTRACT = {
        "mainnet": "0x1390f521A79BaBE99b69B37154D63D431da27A07",
        "testnet": "0xA2eC709125Ea693f5522aEfBBC3cb22fb9146B52",
        # "testnet": "0x87eB0b1B5958c7fD034966925Ea026ad8Bf3d6dD", # old adress used at the time of goerli
    }
    STARK_ABI = {
        "mainnet": starkex_abi_main,
        "testnet": starkex_abi_test,
    }
    POLYGON_ABI = polygon_deposit_abi
    ERC20_ABI = erc20Abi

MAX_INT_ALLOWANCE = '115792089237316195423570985008687907853269984665640564039457584007913129639935'



from typing import TypeVar, Generic, Dict, Tuple, Union, List, Optional
from typing_extensions import TypedDict

T = TypeVar("T")


class Response(TypedDict, Generic[T]):
    status: str
    message: str
    payload: T


class Ticker(TypedDict):
    at: str
    avg_price: str
    high: str
    last: str
    low: str
    open: str
    price_change_percent: str
    volume: str
    amount: str


class MarketType(TypedDict):
    at: str
    ticker: Ticker


FullDayPricePayload = Dict[str, MarketType]

CandleStickPayload = Tuple[int, int, int, int, int, int]


class AskBid(TypedDict):
    id: int
    uuid: str
    side: str
    ord_type: str
    price: str
    avg_price: str
    state: str
    market: str
    created_at: str
    updated_at: str
    origin_volume: str
    remaining_volume: str
    executed_volume: str
    maker_fee: str
    taker_fee: str
    trades_count: int


class OrderBookPayload(TypedDict):
    asks: List[AskBid]
    bids: List[AskBid]


class RecentTradesPayload(TypedDict):
    id: int
    price: int
    amount: int
    total: int
    market: str
    created_at: int
    taker_type: str


class CreateOrderNoncePayload(TypedDict):
    nonce: int
    msg_hash: str


class Signature(TypedDict):
    r: str
    s: str


class CreateNewOrderBody(CreateOrderNoncePayload):
    signature: Signature

class CreateOrderNonceBody(TypedDict):
    market: str
    ord_type: str
    price: float
    side: str
    volume: float

class Order(TypedDict):
    id: int
    uuid: str
    side: str
    ord_type: str
    price: str
    avg_price: str
    state: str
    market: str
    created_at: str
    updated_at: str
    origin_volume: str
    remaining_volume: str
    executed_volume: str
    maker_fee: str
    taker_fee: str
    trades_count: int


class OrderPayload(Order):
    trades: list


class CancelOrder(Order):
    order_id: int


class TradePayload(TypedDict):
    id: int
    price: str
    amount: str
    total: str
    fee_currency: str
    fee: str
    fee_amount: str
    market: str
    created_at: str
    taker_type: str
    side: str
    order_id: int


class LoginPayload(TypedDict):
    uid: str


class TokenType(TypedDict):
    refresh: str
    access: str


class LoginResponse(Response[LoginPayload]):
    token: TokenType


class ProfileInformationPayload(TypedDict):
    name: str
    customer_id: str
    img: str
    username: str
    stark_key: str


class Balance(TypedDict):
    currency: str
    balance: str
    locked: str
    deposit_address: str


class ProfitAndLoss(TypedDict):
    currency: str
    pnl_currency: str
    total_credit: str
    total_debit: str
    total_credit_value: str
    total_debit_value: str
    average_buy_price: str
    average_sell_price: str
    average_balance_price: str
    total_balance_value: str


ProfitAndLossPayload = List[ProfitAndLoss]

class CoinStat(TypedDict):
    stark_asset_id: str
    quanitization: str
    token_contract: str
    decimal: str
    symbol: str
    blockchain_decimal: str

CoinStatPayload = Dict[str, CoinStat]

class ListDepositParams(TypedDict):
    limit: int
    page: int
    network: str

class ListWithdrawalParams(TypedDict):
    page: Optional[int]
    network: Optional[str]

class InitiateWithdrawalPayload(TypedDict):
    amount: float
    symbol: str
    network: Optional[str]

class StarkSignature(TypedDict):
    r: str
    s: str
    recoveryParam: Optional[int]

class ProcessFastWithdrawalPayload(TypedDict):
    msg_hash: str
    signature: StarkSignature
    fastwithdrawal_withdrawal_id: int

class InternalTransferKey(TypedDict):
    organization_key: str
    api_key: str

class InternalTransferInitiateBody(InternalTransferKey):
    client_reference_id: Optional[int]
    currency: str
    amount: float
    destination_address: str

class InternalTransferProcessBody(InternalTransferKey):
    signature: StarkSignature
    nonce: int
    msg_hash: str

class ListInternalTransferParams(TypedDict):
    limit: Optional[int]
    offset: Optional[int]



import asyncio
import websockets.client
import websockets.exceptions
from typing import Optional, Literal
from .exception import AuthenticationError, ConnectNotCalled
import json
import socket


class WsClient:
    def __init__(self, type: str = 'public',
                 option: Literal['mainnet', 'testnet'] = 'mainnet', jwt: Optional[str] = None):
        base_url = "wss://api.tanx.fi" if option == 'mainnet' else 'wss://api-testnet.tanx.fi'
        self.websocket: Optional[websockets.client.WebSocketClientProtocol] = None
        self.connection: str
        if type == 'public':
            self.connection = f"{base_url}/public"
        else:
            if not jwt:
                raise AuthenticationError(
                    'JWT token must be provided for private connections')
            self.connection = f"{base_url}/private?auth_header={jwt}"

    async def connect(self):
        self.websocket = await websockets.client.connect(self.connection)

    async def disconnect(self):
        if self.websocket is not None:
            try:
                await self.websocket.close()
            except socket.gaierror:
                print("Socket gaia error, let's disconnect anyway...")
            except websockets.exceptions.ConnectionClosedError:
                print("WebSockets connection closed error, let's disconnect anyway...")
            except websockets.exceptions.ConnectionClosedOK:
                print("WebSockets connection closed ok, let's disconnect anyway...")
            except ConnectionResetError:
                print("Connection reset error, let's disconnect anyway...")
            del self.websocket

    async def _send(self, data: dict) -> None:
        if not self.websocket:
            raise ConnectNotCalled("wsclient.connect() was not called")
        while not self.websocket:
            await asyncio.sleep(0.1)
        try:
            await self.websocket.send(json.dumps(data))
        except socket.gaierror:
            print("Socket gaia error, message not sent...")
        except websockets.exceptions.ConnectionClosedError:
            print("WebSockets connection closed error, message not sent...")
        except websockets.exceptions.ConnectionClosedOK:
            print("WebSockets connection closed ok, message not sent...")
        except ConnectionResetError:
            print("Connection reset error, message not sent...")

    async def subscribe(self, streams: "list[str]"):
        data = {
            "event": "subscribe",
            "streams": streams
        }
        await self._send(data)

    async def unsubscribe(self, streams: "list[str]"):
        data = {
            "event": "unsubscribe",
            "streams": streams
        }
        await self._send(data)
